import re
import string
from functools import reduce
from draughts.core.move import to_variant
from draughts import Game, Move


class _PDNGame:
    def __init__(self, pdn_text):
        self._tag_value_to_int = ["WhiteRating", "BlackRating"]
        self.values_to_variant = {20: "standard", 21: "english", 22: "italian", 23: "american pool", 24: "spanish", 25: "russian", 26: "brazilian", 27: "canadian", 28: "portuguese", 29: "czech", 30: "turkish", 31: "thai", 40: "frisian", 41: "spantsiretti"}
        self.tags = {}
        self.moves = []
        self.variant = None
        self.notation = None
        self.notation_type = None
        self.game_ending = '*'
        self.pdn_text = pdn_text
        self._rest_of_games = []
        self._read()

    def _read(self):
        lines = self.pdn_text
        lines = lines.split('\n')
        tag_lines = []
        last_tag_line = -1
        for index, line in enumerate(lines):
            if line.startswith('['):
                tag_lines.append(line)
                last_tag_line = index
            elif re.sub(r'\s', '', line):
                break

        for tag_line in tag_lines:
            line = tag_line[1:-1]
            quote_index = line.index('"')
            name = line[:quote_index - 1]
            value = line[quote_index + 1:-1]
            if name in self._tag_value_to_int:
                value = int(value)
            self.tags[name] = value

        rest_of_games = []
        last_move_line = -1
        move_lines = []
        for index, line in enumerate(lines[last_tag_line + 1:]):
            splitted_line = re.split(r'\s(1-0|1/2-1/2|0-1|2-0|1-1|0-2|0-0|\*)\s', ' ' + line + ' ', maxsplit=1)
            if len(splitted_line) == 3:
                move_lines.append(splitted_line[0])
                last_move_line = index
                self.game_ending = splitted_line[1]
                rest_of_games.append(splitted_line[2])
                break
            if re.sub(r'\s', '', line):
                move_lines.append(line)
                last_move_line = index

        rest_of_games += lines[last_tag_line + 1 + last_move_line + 1:]

        moves = " ".join(move_lines)

        # Changes to the PDN

        # From https://stackoverflow.com/a/37538815/10014873
        def remove_text_between_parens(text):
            n = 1  # run at least once
            while n:
                text, n = re.subn(r'\([^()]*\)', '', text)  # remove non-nested/flat balanced parts
            return text

        def remove_text_between_brackets(text):
            n = 1  # run at least once
            while n:
                text, n = re.subn(r'{[^{}]*}', '', text)  # remove non-nested/flat balanced parts
            return text

        moves = remove_text_between_parens(moves)
        moves = remove_text_between_brackets(moves)
        moves = re.sub(r" +", " ", moves)
        moves = re.sub(r"\$[0-9]+", "", moves)
        moves = moves.replace('. ...', '...')
        moves = moves.replace('...', '.')
        moves = moves.replace('?', '')
        moves = moves.replace('!', '')
        moves = moves.replace('. ', '.')
        moves = moves.replace('- ', '-')
        moves = moves.replace('x ', 'x')
        moves = moves.replace(': ', ':')

        move_numbers = re.findall(r"\d+\.", moves)
        double_numbers = list(set(filter(lambda move: move_numbers.count(move) >= 2, move_numbers)))
        for move_number in double_numbers:
            moves = moves[:moves.index(move_number) + len(move_number)] + moves[moves.index(move_number) + len(move_number):].replace(move_number, "")

        moves = moves.split(".")[1:]
        if not moves:
            return
        starts = self.tags.get('FEN', 'W')
        if starts.startswith('W'):
            moves = list(reduce(lambda x, y: x + y.split()[:2], moves, []))
        else:
            moves = [moves[0].split()[0]] + list(reduce(lambda x, y: x + y.split()[:2], moves[1:], []))
        results = ["1-0", "1/2-1/2", "0-1", "2-0", "1-1", "0-2", "0-0", "*"]
        moves = moves[:-1] if moves[-1] in results else moves
        self.moves = moves

        if "GameType" in self.tags:
            game_type = self.tags["GameType"]
            values = game_type.split(',')
            variant_number = int(values[0])
            self.variant = self.values_to_variant.get(variant_number, None)
            if len(values) == 5:
                notation = values[4]
                self.notation_type = notation[0].lower()
                self.notation = int(notation[1])
        else:  # Try to guess the variant
            board_10 = ['31', '32', '33', '34', '35', '16', '17', '18', '19', '20']
            board_8 = ['21', '22', '23', '24', '9', '09', '10', '11', '12']
            first_move = moves[0]
            if list(filter(lambda move: move.startswith(first_move), board_10)):
                self.variant = "standard"
            elif list(filter(lambda move: move.startswith(first_move), board_8)):
                self.variant = "english"
            elif first_move[0] in string.ascii_letters:
                self.variant = "russian"

        self._rest_of_games = rest_of_games

    def get_titles(self):
        return [self.tags.get("WhiteTitle", ""), self.tags.get("BlackTitle", "")]

    def get_ratings(self):
        return [self.tags.get("WhiteRating", ""), self.tags.get("BlackRating", "")]

    def get_na(self):
        return [self.tags.get("WhiteNA", ""), self.tags.get("BlackNA", "")]

    def get_types(self):
        return [self.tags.get("WhiteType", ""), self.tags.get("BlackType", "")]

    def _get_rest_of_games(self):
        return '\n'.join(self._rest_of_games)


class PDNReader:
    def __init__(self, pdn_text=None, filename=None, encodings=None):
        if encodings is None:
            encodings = ['utf8', 'ISO 8859/1']
        if type(encodings) == str:
            encodings = [encodings]
        assert pdn_text or filename
        if not pdn_text:
            pdn_text = ''
            for encoding in encodings:
                try:
                    with open(filename, encoding=encoding) as pdn_file:
                        pdn_text = pdn_file.read()
                    break
                except Exception:
                    pass
        self.pdn_text = pdn_text
        self.pdn_text = re.sub('\n +', '\n', self.pdn_text)
        self.pdn_text = re.sub('\n\n+', '\n\n', self.pdn_text)
        self.games = []
        game = _PDNGame(self.pdn_text)
        self.games.append(game)
        more_games = game._get_rest_of_games()
        while re.sub(r'\s', '', more_games):
            game = _PDNGame(more_games)
            self.games.append(game)
            more_games = game._get_rest_of_games()


class PDNWriter:
    VARIANT_TO_GAMETYPE = {'standard': 20, 'english': 21, 'italian': 22, 'russian': 25, 'brazilian': 26, 'turkish': 30, 'frisian': 40, 'frysk!': 40}
    SHORT_TO_LONG_GAMETYPE = {'20': '20,W,10,10,N2,0', '21': '21,B,8,8,N1,0', '22': '22,W,8,8,N2,1', '25': '25,W,8,8,A0,0', '26': '26,W,8,8,A0,0', '30': '30,W,8,8,A0,0', '40': '40,W,10,10,N2,0'}

    def __init__(self, filename=None, board=None, moves=None, variant=None, starting_fen=None, tags=None, game_ending='*', replay_moves_from_board=True, file_encoding='utf8', file_mode='a'):
        """
        :param replay_moves_from_board: The already saved pdn_move in move_stack may be wrong because it is pseudolegal
        and doesn't account for ambiguous moves. If replay_moves_from_board is enabled, it will replay all the moves to
        find the correct representation of them.
        """
        assert board or moves
        self.pdn_text = ''
        self.notation_type = None
        self.notation = None

        self.board = board
        if self.board:
            self.moves = self.board.move_stack
            self.variant = self.board.variant
            self.starting_fen = self.board.initial_fen
            self.tags = {}
            self.to_standard_notation = True
        else:
            self.moves = moves
            self.variant = variant or 'standard'
            self.starting_fen = starting_fen or self._startpos_to_fen('startpos')
            self.tags = tags or {}
            self.to_standard_notation = False
        self.game_ending = game_ending

        self.replay_moves_from_board = replay_moves_from_board
        self.filename = filename
        self.file_encoding = file_encoding
        self.file_mode = file_mode
        self._fix_ambiguous_moves()
        self._write()

    def _fix_ambiguous_moves(self):
        if self.moves and type(self.moves[0]) == str:
            return
        game = Game(self.variant, self.starting_fen)
        correct_moves = []
        for move in self.moves:
            correct_move = Move(game, board_move=move.board_move)
            correct_moves.append(correct_move)
            for semi_move in move.board_move:
                game.move(semi_move)
        self.moves = correct_moves

    def _write(self):
        pdn_text = ''
        if 'GameType' not in self.tags:
            short_gametype = str(self.VARIANT_TO_GAMETYPE.get(self.variant.lower(), 20))
            long_gametype = self.SHORT_TO_LONG_GAMETYPE[short_gametype]
            self.tags['GameType'] = long_gametype
        if 'FEN' not in self.tags:
            self.tags['FEN'] = self.starting_fen
        for tag in self.tags:
            pdn_text += f'[{tag} "{self.tags[tag]}"]\n'
        pdn_text += '\n'

        if self.to_standard_notation:
            game_type = self.SHORT_TO_LONG_GAMETYPE.get(self.tags['GameType'], self.tags['GameType'])
            values = game_type.split(',')
            notation = values[4]
            self.notation_type = notation[0].lower()
            self.notation = int(notation[1])

        standard_moves = []
        for move in self.moves:
            if type(move) != str:
                standard_move = move.pdn_move
            else:
                standard_move = move

            if self.to_standard_notation:
                standard_move = to_variant(standard_move, variant=self.variant)

            standard_moves.append(standard_move)

        if not standard_moves:
            pass
        elif len(standard_moves) == 1 and self.starting_fen[0] == 'W':
            pdn_text += f'1. {standard_moves[0]}'
            standard_moves = []
        else:
            if self.starting_fen[0] == 'W':
                pdn_text += f'1. {standard_moves[0]} {standard_moves[1]}'
                standard_moves = standard_moves[2:]
            else:
                pdn_text += f'1... {standard_moves[0]}'
                standard_moves = standard_moves[1:]

        write_move_number = True
        move_number = 2
        while standard_moves:
            move_number_to_write = f'{move_number}. ' if write_move_number else ''
            pdn_text += f' {move_number_to_write}{standard_moves.pop(0)}'
            write_move_number = not write_move_number
            if write_move_number:
                move_number += 1
        pdn_text += f' {self.game_ending}'
        pdn_text += '\n\n'

        self.pdn_text = pdn_text

        if self.filename:
            with open(self.filename, self.file_mode, encoding=self.file_encoding) as file:
                file.write(self.pdn_text)

    def _startpos_to_fen(self, fen):
        if fen != 'startpos':
            return fen
        if self.variant == 'frysk!':
            return 'W:W46,47,48,49,50:B1,2,3,4,5'
        elif self.variant == 'turkish':
            return 'W:W41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56:B9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24'
        elif self.variant in ['brazilian', 'russian', 'english', 'italian']:
            return 'W:W21,22,23,24,25,26,27,28,29,30,31,32:B1,2,3,4,5,6,7,8,9,10,11,12'
        else:
            return 'W:W31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50:B1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20'



import string
import re
from math import ceil


class Move:
    def __init__(self, board=None, board_move=None, hub_move=None, hub_position_move=None, pdn_move=None, pdn_position_move=None, steps_move=None, li_api_move=None, li_one_move=None, has_captures=None, possible_moves=None, possible_captures=None, hub_to_pdn_pseudolegal=False, variant=None, notation=2, squares_per_letter=4):
        self.board_move = board_move
        self.hub_move = hub_move
        self.hub_position_move = hub_position_move
        self.pdn_move = pdn_move
        self.pdn_position_move = pdn_position_move
        self.steps_move = steps_move
        self.li_api_move = li_api_move
        self.li_one_move = li_one_move
        self.ambiguous = None
        self.captures = None
        self.has_captures = has_captures
        self.possible_moves = possible_moves
        self.possible_captures = possible_captures
        self.hub_to_pdn_pseudolegal = hub_to_pdn_pseudolegal
        self.variant = variant
        self.original_pdn_move = self.pdn_move
        self.notation = notation
        self.squares_per_letter = squares_per_letter

        if board_move or hub_move or hub_position_move or pdn_move or pdn_position_move or steps_move or li_api_move or li_one_move:
            if board or possible_moves and possible_captures:
                if not possible_moves or not possible_captures:
                    self.possible_moves, self.possible_captures = board.legal_moves()
                self._to_board()
                self.captures = self.possible_captures[self.possible_moves.index(self.board_move)]
                self.captures = [] if self.captures[0] is None else self.captures
                self.has_captures = bool(self.captures)
                self._from_board()
            else:
                self._no_board()

    def _make_len_2(self, move):
        return f'0{move}' if len(str(move)) == 1 else str(move)

    def _sort_captures(self, captures):
        """
        This function is because hub engines returns the captures in alphabetical order
        (e.g. for the move 231201 scan returns 23x01x07x18 instead of 23x01x18x07)
        """
        if captures and captures[0] is None:
            captures = []
        captures = list(map(self._make_len_2, captures))
        captures.sort()
        captures = ''.join(captures)
        return captures

    def _to_board(self):
        if self.hub_move:
            if "-" in self.hub_move:
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("-"))))
            else:
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("x"))))
        if self.pdn_move:
            self.pdn_move = from_variant(self.pdn_move, variant=self.variant, notation=self.notation, squares_per_letter=self.squares_per_letter)

            if "-" in self.pdn_move:
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("-"))))
            else:
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("x"))))

        if self.hub_position_move:
            moves_li_board = {}
            for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(
                    possible_move[-1][1]) + self._sort_captures(possible_capture)
                moves_li_board[li_move] = possible_move
            board_move = moves_li_board[self.hub_position_move]
            self.board_move = board_move

        elif self.pdn_position_move:
            moves_li_board = {}
            if len(self.pdn_position_move) == 4:
                self.ambiguous = False
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    li_move = self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1])
                    moves_li_board[li_move] = possible_move
            else:
                self.ambiguous = True
                for possible_move, possible_capture in zip(self.possible_moves, self.possible_captures):
                    steps = [possible_move[0][0]]
                    for move in possible_move:
                        steps.append(move[1])
                    li_move = "".join(list(map(self._make_len_2, steps)))
                    moves_li_board[li_move] = possible_move
            board_move = moves_li_board[self.pdn_position_move]
            self.board_move = board_move

        elif self.steps_move:
            board_move = []
            for index in range(1, len(self.steps_move)):
                board_move.append([self.steps_move[index - 1], self.steps_move[index]])
            self.board_move = board_move

        elif self.li_api_move:
            board_move = []
            for move in self.li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])
            self.board_move = board_move

        elif self.li_one_move:
            steps = [int(self.li_one_move[i:i + 2]) for i in range(0, len(self.li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])
            self.board_move = board_move

    def _from_board(self):
        if not self.steps_move:
            steps = [self.board_move[0][0]]
            for move in self.board_move:
                steps.append(move[1])
            self.steps_move = steps

        if not self.hub_position_move:
            self.hub_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])
            if self.captures:
                sorted_captures = self._sort_captures(self.captures)
                sorted_captures = [sorted_captures[i:i + 2] for i in range(0, len(sorted_captures), 2)]
                self.hub_position_move += "".join(list(map(self._make_len_2, sorted_captures)))

        if not self.hub_move:
            positions = [self.hub_position_move[i:i + 2] for i in range(0, len(self.hub_position_move), 2)]
            separator = "x" if self.captures else "-"
            self.hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if not self.pdn_position_move:
            starts_endings = []
            for possible_move in self.possible_moves:
                starts_endings.append(self._make_len_2(possible_move[0][0]) + self._make_len_2(possible_move[-1][1]))
            self.ambiguous = starts_endings.count(self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])) >= 2
            if self.ambiguous:
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.steps_move)))
            else:
                self.pdn_position_move = self._make_len_2(self.board_move[0][0]) + self._make_len_2(self.board_move[-1][1])

        if not self.pdn_move:
            positions = [self.pdn_position_move[i:i + 2] for i in range(0, len(self.pdn_position_move), 2)]
            separator = "x" if self.captures else "-"
            self.pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if not self.li_api_move:
            li_api_move = []
            for move in self.board_move:
                li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))
            self.li_api_move = li_api_move

        if not self.li_one_move:
            self.li_one_move = "".join(list(map(self._make_len_2, self.steps_move)))

    def _no_board(self):
        """
        Makes as many conversions as possible without the board.
        """

        # Board related moves

        if self.steps_move:
            board_move = []
            for index in range(1, len(self.steps_move)):
                board_move.append([self.steps_move[index - 1], self.steps_move[index]])
            self.board_move = board_move

        elif self.li_api_move:
            board_move = []
            for move in self.li_api_move:
                board_move.append([int(move[:2]), int(move[2:])])
            self.board_move = board_move

        elif self.li_one_move:
            steps = [int(self.li_one_move[i:i + 2]) for i in range(0, len(self.li_one_move), 2)]
            board_move = []
            for index in range(1, len(steps)):
                board_move.append([steps[index - 1], steps[index]])
            self.board_move = board_move

        if self.board_move:
            # steps_move
            if not self.steps_move:
                steps = [self.board_move[0][0]]
                for move in self.board_move:
                    steps.append(move[1])
                self.steps_move = steps

            # li_api_move
            if not self.li_api_move:
                li_api_move = []
                for move in self.board_move:
                    li_api_move.append(self._make_len_2(move[0]) + self._make_len_2(move[1]))
                self.li_api_move = li_api_move

            # li_one_move
            if not self.li_one_move:
                self.li_one_move = "".join(list(map(self._make_len_2, self.steps_move)))

        # Hub related moves

        if self.hub_move:
            if "-" in self.hub_move:
                self.has_captures = False
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("-"))))
            else:
                self.has_captures = True
                self.hub_position_move = "".join(list(map(self._make_len_2, self.hub_move.split("x"))))
        elif self.hub_position_move:
            positions = [self.hub_position_move[i:i + 2] for i in range(0, len(self.hub_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            self.hub_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if self.hub_to_pdn_pseudolegal:
            self.pdn_position_move = self.hub_position_move[:4]
            separator = "x" if self.has_captures else "-"
            self.pdn_move = self.pdn_position_move[:2] + separator + self.pdn_position_move[2:]

        # PDN related moves

        if self.pdn_move:
            self.pdn_move = from_variant(self.pdn_move, variant=self.variant, notation=self.notation, squares_per_letter=self.squares_per_letter)

            if "-" in self.pdn_move:
                self.has_captures = False
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("-"))))
            else:
                self.has_captures = True
                self.pdn_position_move = "".join(list(map(self._make_len_2, self.pdn_move.split("x"))))
        elif self.pdn_position_move:
            positions = [self.pdn_position_move[i:i + 2] for i in range(0, len(self.pdn_position_move), 2)]
            separator = "x" if self.has_captures else "-"
            self.pdn_move = separator.join(list(map(lambda position: str(int(position)), positions)))

        if self.pdn_position_move:
            self.ambiguous = not (len(self.pdn_position_move) == 4)


def _rotate_move(internal_move, notation=None, variant=None):
    separators = ['-', 'x', ':']
    splitted_move = None
    correct_seperator = None
    for separator in separators:
        splitted_move = internal_move.split(separator)
        if splitted_move[0] != internal_move:
            correct_seperator = separator
            break
    splitted_move = list(map(int, splitted_move))
    variant_to_notation = {'standard': 2, 'english': 0, 'italian': 2, 'russian': 0, 'brazilian': 0, 'turkish': 0, 'frisian': 2, 'frysk!': 2, 'antidraughts': 2, 'breakthrough': 2}
    value = variant_to_notation.get(variant, 2)
    if notation is None:
        notation = value

    def reverse_column(splitted_int_move):
        per_row = 5
        if variant in ['english', 'italian', 'russian', 'brazilian']:
            per_row = 4
        elif variant == 'turkish':
            per_row = 8
        for index, square in enumerate(splitted_int_move):
            square_in_row = square % per_row
            if square_in_row == 0:
                square_in_row += per_row
            splitted_int_move[index] = ((square - 1) // per_row) * per_row + (per_row - (square_in_row - 1))
        return splitted_int_move

    def reverse_row_and_column(splitted_int_move):
        if variant in ['brazilian', 'russian', 'english', 'italian']:
            squares = 32
        elif variant == 'turkish':
            squares = 64
        else:
            squares = 50
        splitted_int_move = list(map(lambda square: squares + 1 - square, splitted_int_move))
        return splitted_int_move

    def reverse_row(splitted_int_move):
        return reverse_column(reverse_row_and_column(splitted_int_move))

    rotated_move = None
    if notation == 0:
        rotated_move = reverse_row(splitted_move)
    elif notation == 1:
        rotated_move = reverse_row_and_column(splitted_move)
    elif notation == 2:
        rotated_move = splitted_move
    elif notation == 3:
        rotated_move = reverse_column(splitted_move)

    rotated_move = list(map(str, rotated_move))
    return correct_seperator.join(rotated_move)


def _algebraic_to_number(algebraic_move, squares_per_letter=None, variant=None, every_other_square=None):
    if every_other_square is None:
        if variant == 'turkish':
            every_other_square = False
        else:
            every_other_square = True
    algebraic_notation = algebraic_move[0] in string.ascii_letters
    if not algebraic_notation:
        return algebraic_move
    algebraic_move = algebraic_move.lower()
    if squares_per_letter is None:
        squares_per_letter = 5
        if variant in ['english', 'italian', 'russian', 'brazilian']:
            squares_per_letter = 4
        elif variant == 'turkish':
            squares_per_letter = 8

    separators = ['-', 'x', ':']
    special_seperators = ['([a-zA-z]+\d+)']
    splitted_move = None
    correct_seperator = None
    for separator in separators:
        splitted_move = list(filter(bool, re.split(separator, algebraic_move)))
        if splitted_move[0] != algebraic_move:
            correct_seperator = separator
            break

    if not correct_seperator:
        for separator in special_seperators:
            splitted_move = list(filter(bool, re.split(separator, algebraic_move)))
            if splitted_move[0] != algebraic_move:
                correct_seperator = '-'
                break

    numeric_move = []
    for move in splitted_move:
        numeric_move.append(_algebraic_to_numeric_square(move, squares_per_letter, every_other_square=every_other_square))
    numeric_move = list(map(str, numeric_move))
    return correct_seperator.join(numeric_move)


def _algebraic_to_numeric_square(square, squares_per_letter, every_other_square=True):
    algebraic_notation = square[0] in string.ascii_letters
    if not algebraic_notation:
        return square
    if not every_other_square:
        return (int(square[1]) - 1) * squares_per_letter + string.ascii_lowercase.index(square[0]) + 1
    return (int(square[1]) - 1) * squares_per_letter + ceil(string.ascii_lowercase.index(square[0]) // 2) + 1


def _number_to_algebraic(number_move, width=None, variant=None, every_other_square=None):
    if every_other_square is None:
        if variant == 'turkish':
            every_other_square = False
        else:
            every_other_square = True
    algebraic_notation = number_move[0] in string.ascii_letters
    if algebraic_notation:
        return number_move
    if width is None:
        width = 5
        if variant in ['english', 'italian', 'russian', 'brazilian']:
            width = 4
        elif variant == 'turkish':
            width = 8

    separators = ['-', 'x', ':']
    special_seperators = ['([a-zA-z]+\d+)']
    splitted_move = None
    correct_seperator = None
    for separator in separators:
        splitted_move = list(filter(bool, re.split(separator, number_move)))
        if splitted_move[0] != number_move:
            correct_seperator = separator
            break

    if not correct_seperator:
        for separator in special_seperators:
            splitted_move = list(filter(bool, re.split(separator, number_move)))
            if splitted_move[0] != number_move:
                correct_seperator = '-'
                break

    algebraic_move = []
    for move in splitted_move:
        algebraic_move.append(_numeric_to_algebraic_square(move, width, every_other_square=every_other_square))
    algebraic_move = list(map(str, algebraic_move))
    return correct_seperator.join(algebraic_move)


def _numeric_to_algebraic_square(square, width, every_other_square=True):
    algebraic_notation = square[0] in string.ascii_letters
    if algebraic_notation:
        return square
    square = int(square)
    row = ceil(square / width) - 1
    column = (square - 1) % width
    if every_other_square:
        column *= 2
        column += 1 if row % 2 == 1 else 0
    return string.ascii_lowercase[column] + str(row + 1)


def from_variant(move, variant=None, notation=None, squares_per_letter=None):
    variant = variant.lower() if variant else variant
    move = _algebraic_to_number(move, variant=variant, squares_per_letter=squares_per_letter)
    move = _rotate_move(move, variant=variant, notation=notation)
    return move


def to_variant(move, variant=None, notation=None, width=None, every_other_square=None, to_algebraic=None):
    variant = variant.lower() if variant else variant
    move = _rotate_move(move, variant=variant, notation=notation)
    if to_algebraic or variant in ['russian', 'brazilian', 'turkish']:
        move = _number_to_algebraic(move, variant=variant, width=width, every_other_square=every_other_square)
    return move



from draughts.engines.dxp import DXPEngine
from draughts.engines.hub import HubEngine
from draughts.engines.checkerboard import CheckerBoardEngine


class Limit:
    def __init__(self, time=None, inc=None, depth=None, nodes=None, movetime=None):
        assert time is not None or depth is not None or nodes is not None or movetime is not None
        self.time = time
        self.inc = inc
        self.depth = depth
        self.nodes = nodes
        self.movetime = movetime


class PlayResult:
    def __init__(self, move=None, ponder=None, info=None, draw_offered=False, resign=False):
        self.move = move
        self.ponder = ponder
        self.info = info
        self.draw_offered = draw_offered
        self.resign = resign



import draughts.engines.dxp_communication.dxp_run as dxp
import draughts
import draughts.engine
import subprocess
import os
import signal
import threading
import time
import logging
from importlib import reload

logger = logging.getLogger(__name__)


class DXPEngine:
    def __init__(self, command, options=None, initial_time=0, ENGINE=5):
        global dxp
        dxp = reload(dxp)
        if options is None:
            options = {}
        self.initial_time = initial_time
        self.max_moves = 0
        self.ip = '127.0.0.1'
        self.port = '27531'
        self.engine_opened = True  # Whether the engine is already open or pydraughts should open it
        self.wait_to_open_time = 10
        if type(command) == str:
            command = [command]
        self.command = command
        self.ENGINE = ENGINE
        self.info = {}
        self.id = {}
        self.console = dxp.tConsoleHandler
        self.receiver = dxp.tReceiveHandler
        self.game_started = False
        self._last_move = None

        for name, value in options.items():
            self.setoption(name, value)

        if not self.engine_opened:
            cwd = os.path.realpath(os.path.expanduser("."))
            command = list(filter(bool, self.command))
            command = " ".join(command)
            self.p = self._open_process(command, cwd)
            self.thr = threading.Thread(target=self._recv)
            self.thr.start()

        self.start_time = time.perf_counter_ns()

    def setoption(self, name, value):
        if name == 'engine-opened':
            self.engine_opened = value
        elif name == 'ip':
            self.ip = str(value)
        elif name == 'port':
            self.port = str(value)
        elif name == 'wait-to-open-time':
            self.wait_to_open_time = int(value)
        elif name == 'max-moves':
            self.max_moves = 0
        elif name == 'initial-time':
            self.initial_time = 0

    def _open_process(self, command, cwd=None, shell=True, _popen_lock=threading.Lock()):
        kwargs = {
            "shell": shell,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": subprocess.PIPE,
            "bufsize": 1,  # Line buffered
            "universal_newlines": True,
        }
        logger.debug(f'command: {command}, cwd: {cwd}')

        if cwd is not None:
            kwargs["cwd"] = cwd

        # Prevent signal propagation from parent process
        try:
            # Windows
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        except AttributeError:
            # Unix
            kwargs["preexec_fn"] = os.setpgrp

        with _popen_lock:  # Work around Python 2 Popen race condition
            return subprocess.Popen(command, **kwargs)

    def kill_process(self):
        if not self.engine_opened:
            try:
                # Windows
                self.p.send_signal(signal.CTRL_BREAK_EVENT)
            except AttributeError:
                # Unix
                os.killpg(self.p.pid, signal.SIGKILL)

            self.p.communicate()

    def _connect(self):
        if not self.engine_opened:
            wait_time = self.start_time / 1e9 + self.wait_to_open_time - time.perf_counter_ns() / 1e9
            logger.debug(f'wait time: {wait_time}')
            if wait_time > 0:
                time.sleep(wait_time)
        self.console.run_command(f'connect {self.ip} {self.port}')

    def _start(self, board, time):
        self._connect()
        color = 'B' if board.whose_turn() == draughts.WHITE else 'W'
        time = str(round(time // 60))
        moves = self.max_moves
        self.console.run_command(f'setup {board.initial_fen} {board.variant}')
        self.console.run_command(f'gamereq {color} {time} {moves}')
        accepted = self._recv_accept()
        logger.debug(f'Aceepted: {accepted}')
        self.id["name"] = dxp.current.engineName

    def _recv(self):
        while True:
            line = self.p.stdout.readline()

            line = line.rstrip()

            if line:
                logging.debug(f"{self.ENGINE} %s >> %s {self.p.pid} {line}")

    def _recv_accept(self):
        while True:
            if dxp.accepted is not None:
                return dxp.accepted

    def _recv_move(self):
        while True:
            if not dxp.tReceiveHandler.isListening:
                break
            if self._last_move != dxp.last_move:
                self._last_move = dxp.last_move
                logger.debug(f'new last move: {self._last_move}')
                return self._last_move

    def play(self, board):
        if not self.game_started:
            self._start(board, self.initial_time)
            self.game_started = True
        if board.move_stack:
            move = board.move_stack[-1].li_one_move
            move = [move[i:i+2] for i in range(0, len(move), 2)]
            move = '-'.join(move)
            self.console.run_command(f'sm {move}')
        best_move = self._recv_move()
        if best_move:
            best_move = draughts.Move(board, best_move)
        return draughts.engine.PlayResult(best_move, None)

    def quit(self):
        self.console.run_command('gameend 0')



# This file is an adaptation of DXC100_draughts_client (https://github.com/akalverboer/DXC100_draughts_client) by akalverboer which is licensed under the MIT License.

import socket
import logging
import draughts

logger = logging.getLogger(__name__)

DXP_WHITE = 0
DXP_BLACK = 1


class GameStatus:
    def __init__(self, fen='startpos', myColor=DXP_WHITE, started=False, variant='standard'):
        self.fen = fen
        self.myColor = myColor
        self.started = started
        self.variant = variant
        self.pos = draughts.Game(fen=fen, variant=variant)
        self.color = self.get_color()
        self.engineName = ''
        self.result = None

    def get_color(self):
        return DXP_WHITE if self.pos.whose_turn() == draughts.WHITE else DXP_BLACK


class MySocket:
    def __init__(self):
        self.sock = None

    def open(self):
        try:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        except Exception:
            self.sock = None
            raise Exception("socket exception: failed to open")
        return self

    def connect(self, host, port):
        self.sock.settimeout(10)  # timeout for connection
        try:
            self.sock.connect((host, port))
        except socket.error as msg:
            self.sock = None
            raise Exception("connection exception: failed to connect")
        if self.sock is not None:
            self.sock.settimeout(None)  # default
        return self

    def send(self, msg):
        try:
            self.sock.send(bytes(msg, 'utf-8') + b"\0")
        except Exception:
            raise Exception("send exception: no connection")
        return None

    def receive(self):
        msg = ""
        while True:
            # Collect message chunks until null character found
            try:
                chunk = self.sock.recv(1024)
            except Exception:
                raise Exception("receive exception: no connection")

            if chunk == "":
                raise Exception("receive exception: socket connection broken")
            msg += chunk.decode()
            if msg.find("\0") > -1:
                break
            if len(msg) > 128:
                break  # too long, no null char

        logger.debug("final msg: " + msg)
        msg = msg.replace("\0", "")  # remove all null chars

        # Use strip to remove all whitespace at the start and end.
        # Including spaces, tabs, newlines and carriage returns.
        msg = msg.strip()
        return msg


class DamExchange:
    def parse(self, msg):
        # Parse incoming DXP message. Returns relevant items depending on mtype.
        result = {}
        mtype = msg[0:1]
        if mtype == "C":  # CHAT
            result['type'] = "C"
            result['text'] = msg[1:127]
        elif mtype == "R":  # GAMEREQ (only received by FOLLOWER)
            result['type'] = "R"
            result['name'] = msg[3:35].strip()  # initiator
            result['fColor'] = msg[35:36]  # color of follower
            result['gameTime'] = msg[36:40]
            result['numMoves'] = msg[40:44]
            result['posInd'] = msg[44:45]
            if result['posInd'] != "A":
                result['mColor'] = msg[45:46]  # color to move for position
                result['pos'] = msg[46:96]
        elif mtype == "A":  # GAMEACC
            result['type'] = "A"
            result['engineName'] = msg[1:33].strip()  # follower name
            result['accCode'] = msg[33:34]
        elif mtype == "M":  # MOVE
            result['type'] = "M"
            result['time'] = msg[1:5]
            result['from'] = msg[5:7]
            result['to'] = msg[7:9]
            result['nCaptured'] = msg[9:11]
            result['captures'] = []
            for i in range(int(result['nCaptured'])):
                s = i * 2
                result['captures'].append(msg[11 + s:13 + s])
        elif mtype == "E":  # GAMEEND
            result['type'] = "E"
            result['reason'] = msg[1:2]
            result['stop'] = msg[2:3]
        elif mtype == "B":  # BACKREQ
            result['type'] = "B"
            result['moveId'] = msg[1:4]
            result['mColor'] = msg[4:5]
        elif mtype == "K":  # BACKACC
            result['type'] = "K"
            result['accCode'] = msg[1:2]
        else:
            result['type'] = "?"
        return result

    def msg_chat(self, str):
        # Generate CHAT message. Example: CWhat do you think about move 35?
        msg = "C" + str
        return msg

    def msg_gamereq(self, myColor, gameTime, numMoves, pos=None, colorToMove=None):
        # Generate GAMEREQ message. Example: R01Tornado voor Windows 4.0        W060065A
        gamereq = []
        gamereq.append("R")  # header
        gamereq.append("01")  # version

        gamereq.append("DXP Client".ljust(32)[:32])  # iName: fixed length padding spaces
        gamereq.append('Z' if myColor == DXP_WHITE else 'W')  # fColor: color of follower (server)
        gamereq.append(str(gameTime).zfill(3))  # gameTime: time limit of game (ex: 090)
        gamereq.append(str(numMoves).zfill(3))  # numMoves: number of moves of time limit (ex: 050)
        if pos is None or colorToMove is None:
            gamereq.append("A")  # posInd == A: use starting position
        else:
            gamereq.append("B")  # posInd == B: use parameters pos and colorToMove
            gamereq.append("W" if colorToMove == DXP_WHITE else "Z")  # mColor
            gamereq.append(pos.get_dxp_fen())  # board

        msg = ""
        for item in gamereq:
            msg = msg + item
        return msg

    def msg_move(self, steps, captures, timeSpend):
        # Generate MOVE message. Example: M001205250422122320
        # Parm rmove is a "two-color" move
        move = []
        move.append("M")  # header
        move.append(str(timeSpend % 10000).zfill(4))  # mTime: 0000 .. 9999
        move.append(str(steps[0] % 100).zfill(2))  # mFrom
        move.append(str(steps[-1] % 100).zfill(2))  # mTo
        move.append(str(len(captures) % 100).zfill(2))  # mNumCaptured: number of takes (captures)
        for k in captures:
            move.append(str(k % 100).zfill(2))  # mCaptures

        msg = ""
        for item in move:
            msg = msg + item
        return msg

    def msg_gameend(self, reason):
        # Generate GAMEEND message. Example: E00
        gameend = []
        gameend.append("E")  # header
        gameend.append(str(reason)[0])  # reason:  0 > unknown  1 > I lose  2 > draw  3 > I win
        gameend.append("1")  # stop code: 0 > next game preferred  1: > no next game
        msg = ""
        for item in gameend:
            msg = msg + item
        return msg

    def msg_backreq(self, moveId, colorToMove):
        # Generate BACKREQ message. Example: B005Z
        backreq = []
        backreq.append("B")
        backreq.append(str(moveId % 1000).zfill(3))  # moveId
        backreq.append("W" if colorToMove == DXP_WHITE else "Z")  # mColor
        msg = ""
        for item in backreq:
            msg = msg + item
        return msg

    def msg_backacc(self, accCode):
        # Generate BACKREQ message. Example: K1
        backreq = []
        backreq.append("K")
        backreq.append(str(accCode[0]))  # accCode
        msg = ""
        for item in backreq:
            msg = msg + item
        return msg



from functools import reduce

WHITE = 2
BLACK = 1


class BoardSearcher:

    def build(self, board):
        self.board = board
        self.uncaptured_pieces = list(filter(lambda piece: not piece.captured, board.pieces))
        self.open_positions = []
        self.filled_positions = []
        self.player_positions = {}
        self.player_pieces = {}
        self.position_pieces = {}

        self.build_filled_positions()
        self.build_open_positions()
        self.build_player_positions()
        self.build_player_pieces()
        self.build_position_pieces()

    def build_filled_positions(self):
        self.filled_positions = reduce((lambda open_positions, piece: open_positions + [piece.position]), self.uncaptured_pieces, [])

    def build_open_positions(self):
        self.open_positions = [position for position in range(1, self.board.position_count) if not position in self.filled_positions]

    def build_player_positions(self):
        self.player_positions = {
            1: reduce((lambda positions, piece: positions + ([piece.position] if piece.player == BLACK else [])), self.uncaptured_pieces, []),
            2: reduce((lambda positions, piece: positions + ([piece.position] if piece.player == WHITE else [])), self.uncaptured_pieces, [])
        }

    def build_player_pieces(self):
        self.player_pieces = {
            1: reduce((lambda pieces, piece: pieces + ([piece] if piece.player == BLACK else [])), self.uncaptured_pieces, []),
            2: reduce((lambda pieces, piece: pieces + ([piece] if piece.player == WHITE else [])), self.uncaptured_pieces, [])
        }

    def build_position_pieces(self):
        self.position_pieces = {piece.position: piece for piece in self.uncaptured_pieces}

    def get_pieces_by_player(self, player_number):
        return self.player_pieces[player_number]

    def get_positions_by_player(self, player_number):
        return self.player_positions[player_number]

    def get_pieces_in_play(self):
        return self.player_pieces[self.board.player_turn] if not self.board.piece_requiring_further_capture_moves else [self.board.piece_requiring_further_capture_moves]

    def get_piece_by_position(self, position):
        return self.position_pieces.get(position)



import ctypes
import draughts
from math import ceil


def get_board(board):
    row = (ctypes.c_int * 8)
    checkerboard_board = (row * 8)()

    # From CheckerBoard API:
    WHITE = 1
    BLACK = 2
    MAN = 4
    KING = 8
    FREE = 0

    for loc in range(1, board.board.position_count + 1):
        piece = board.board.searcher.get_piece_by_position(loc)
        row = ceil(loc / board.board.width) - 1  # From get_row_from_position
        row = (board.board.height - 1) - row
        column = (loc - 1) % board.board.width  # From get_column
        if row % 2 == 0:
            column = (column + 1) * 2 - 2  # To account for the always empty white squares
        else:
            column = (column + 1) * 2 - 1  # To account for the always empty white squares
        number = FREE
        if piece:
            # In Checkerboard black starts first, so the colors are reversed
            if piece.player == draughts.WHITE and not piece.king:
                number = BLACK + MAN
            elif piece.player == draughts.BLACK and not piece.king:
                number = WHITE + MAN
            elif piece.player == draughts.WHITE and piece.king:
                number = BLACK + KING
            elif piece.player == draughts.BLACK and piece.king:
                number = WHITE + KING
        checkerboard_board[column][row] = number

    return checkerboard_board


def from_board(checker_board, old_board):
    # board_numbers = [[square for square in column] for column in checker_board]
    fen = 'B' if old_board.whose_turn() == draughts.WHITE else 'W'  # switch turns

    # From CheckerBoard API:
    WHITE = 1
    BLACK = 2
    MAN = 4
    KING = 8
    FREE = 0

    for loc in range(1, old_board.board.position_count + 1):
        row = ceil(loc / old_board.board.width) - 1  # From get_row_from_position
        row = (old_board.board.height - 1) - row
        column = (loc - 1) % old_board.board.width  # From get_column
        if row % 2 == 0:
            column = (column + 1) * 2 - 2  # To account for the always empty white squares
        else:
            column = (column + 1) * 2 - 1  # To account for the always empty white squares
        # In Checkerboard black starts first, so the colors are reversed
        if checker_board[column][row] == FREE:
            fen += 'e'
        elif checker_board[column][row] == WHITE + MAN:
            fen += 'b'
        elif checker_board[column][row] == BLACK + MAN:
            fen += 'w'
        elif checker_board[column][row] == WHITE + KING:
            fen += 'B'
        elif checker_board[column][row] == BLACK + KING:
            fen += 'W'
    return fen



import ctypes
import draughts
from draughts.engines.checkerboard_extra.get_checker_board import get_board, from_board
import os

from msl.loadlib import Server32

with open(os.path.join(os.path.dirname(__file__), 'engine_name.txt')) as file:
    DLL_name = file.read()


class Engine32Server(Server32):

    def __init__(self, host, port, **kwargs):
        super(Engine32Server, self).__init__(DLL_name, 'windll', host, port)

    def enginecommand(self, command):
        output = ctypes.create_string_buffer(b'', 1024)
        result = self.lib.enginecommand(ctypes.create_string_buffer(bytes(command.encode('ascii')), 256), output)
        return output.value, result

    def getmove(self, game, maxtime, time, increment, movetime):

        # From CheckerBoard API:
        WHITE = 1
        BLACK = 2

        board = get_board(game)

        # Reversed color because red (black) starts first in Checkerboard and not white
        color = BLACK if game.whose_turn() == draughts.WHITE else WHITE

        info = 0
        moreinfo = 0
        if movetime:
            info = info | (1 << 1)  # 2nd bit means the engine has to think for exactly maxtime seconds
        elif time:
            if time / .01 > 2 ** 15 - 1 or increment / .01 > 2 ** 15 - 1:
                info = info | (1 << 3)  # 0.1 seconds
                info = info | (1 << 4)  # 0.1 seconds
                time = int(time / .1)
                increment = int(increment / .1)
            elif time / .001 > 2 ** 15 - 1 or increment / .001 > 2 ** 15 - 1:
                info = info | (1 << 3)  # 0.01 seconds
                time = int(time / .01)
                increment = int(increment / .01)
            else:
                info = info | (1 << 4)  # 0.001 seconds
                time = int(time / .001)
                increment = int(increment / .001)
            bin_time = bin(time)[2:].zfill(16)
            if len(bin_time) > 16:
                bin_time = '1' * 16
            bin_inc = bin(increment)[2:].zfill(16)
            if len(bin_inc) > 16:
                bin_inc = '1' * 16
            moreinfo = eval('0b' + bin_time + bin_inc)

        if movetime:
            maxtime = ctypes.c_double(float(movetime))
        else:
            maxtime = ctypes.c_double(float(maxtime))
        output = ctypes.create_string_buffer(b'', 1024)
        playnow = ctypes.c_int(0)

        class coor(ctypes.Structure):
            _fields_ = [("x", ctypes.c_int), ("y", ctypes.c_int)]

        class CBmove(ctypes.Structure):
            _fields_ = [("jumps", ctypes.c_int), ("newpiece", ctypes.c_int), ("oldpiece", ctypes.c_int), ("from", coor), ("to", coor), ("path", coor * 12), ("del", coor * 12), ("delpiece", ctypes.c_int * 12)]

        self.lib.getmove.argtypes = [((ctypes.c_int * 8) * 8), ctypes.c_int, ctypes.c_double, (ctypes.c_char * 1024), ctypes.POINTER(ctypes.c_int), ctypes.c_int, ctypes.c_int, ctypes.POINTER(CBmove)]

        cbmove = CBmove()

        result = self.lib.getmove(board, color, maxtime, output, ctypes.byref(playnow), info, moreinfo, ctypes.byref(cbmove))

        old_fen = game.get_fen()
        new_fen = from_board(board, game)
        our_pieces, opponents_pieces = (['w', 'W'], ['b', 'B']) if old_fen[0] == 'W' else (['b', 'B'], ['w', 'W'])
        captures = []
        start_pos, end_pos = None, None
        for index in range(1, len(old_fen)):
            if old_fen[index] in our_pieces and new_fen[index] == 'e':
                start_pos = index
            elif new_fen[index] in our_pieces and old_fen[index] == 'e':
                end_pos = index
            elif old_fen[index] in opponents_pieces and new_fen[index] == 'e':
                captures.append(index)
        if start_pos and end_pos:
            hub_pos_move = game.make_len_2(start_pos) + game.make_len_2(end_pos) + game.sort_captures(captures)
        else:
            hub_pos_move = None

        cbmove_output_2 = {}
        cbmove_output_2['jumps'] = cbmove.jumps
        cbmove_output_2['oldpiece'] = cbmove.oldpiece
        cbmove_output_2['newpiece'] = cbmove.newpiece
        cbmove_output_2['to'] = cbmove.to.x, cbmove.to.y
        cbmove.to = getattr(cbmove, 'from')
        cbmove_output_2['from'] = cbmove.to.x, cbmove.to.y
        cbmove_output_2['path'] = [(cbmove.path[i].x, cbmove.path[i].y) for i in range(12)]
        cbmove.path = getattr(cbmove, 'del')
        cbmove_output_2['del'] = [(cbmove.path[i].x, cbmove.path[i].y) for i in range(12)]
        cbmove_output_2['delpiece'] = [cbmove.delpiece[i] for i in range(12)]
        return hub_pos_move, output.value, cbmove_output_2, result



from draughts.core.board_searcher import BoardSearcher
from draughts.core.board_initializer import BoardInitializer
from functools import reduce
import pickle

WHITE = 2
BLACK = 1


class Board:

    def __init__(self, variant='standard', fen='startpos'):
        if fen != 'startpos':
            self.player_turn = WHITE if fen[0].lower() == 'w' else BLACK
        else:
            self.player_turn = WHITE
        if variant in ['brazilian', 'russian', 'english', 'italian']:
            self.width = 4
            self.height = 8
        elif variant == 'turkish':
            self.width = 8
            self.height = 8
        else:
            self.width = 5
            self.height = 10
        self.position_count = self.width * self.height
        if variant == 'frysk!':
            self.rows_per_user_with_pieces = 1
        elif variant == 'turkish':
            self.rows_per_user_with_pieces = 2
        elif variant in ['brazilian', 'russian', 'english', 'italian']:
            self.rows_per_user_with_pieces = 3
        else:
            self.rows_per_user_with_pieces = 4
        self.position_layout = {}
        self.piece_requiring_further_capture_moves = None
        self.previous_move_was_capture = False
        self.variant = variant
        self.fen = fen
        self.searcher = BoardSearcher()
        BoardInitializer(self, self.fen).initialize()

    def count_movable_player_pieces(self, player_number=1, captures=None):
        if captures is None:
            captures = []
        return reduce((lambda count, piece: count + (1 if piece.is_movable(captures) else 0)), self.searcher.get_pieces_by_player(player_number), 0)

    def get_possible_moves(self, captures):
        capture_moves = self.get_possible_capture_moves(captures)

        return capture_moves if capture_moves else self.get_possible_positional_moves()

    def get_possible_capture_moves(self, captures):
        return reduce((lambda moves, piece: moves + piece.get_possible_capture_moves(captures)), self.searcher.get_pieces_in_play(), [])

    def get_possible_positional_moves(self):
        return reduce((lambda moves, piece: moves + piece.get_possible_positional_moves()), self.searcher.get_pieces_in_play(), [])

    def position_is_open(self, position):
        return not self.searcher.get_piece_by_position(position)

    def create_new_board_from_move(self, move, move_number, captures, return_captured=False):
        new_board = pickle.loads(pickle.dumps(self, -1))  # A lot faster that deepcopy
        enemy_position = None

        if move in self.get_possible_capture_moves(captures):
            if return_captured:
                enemy_position = new_board.perform_capture_move(move, move_number, captures, return_captured=return_captured)
            else:
                new_board.perform_capture_move(move, move_number, captures)
        else:
            new_board.perform_positional_move(move, move_number)

        if return_captured:
            return new_board, enemy_position
        else:
            return new_board

    def push_move(self, move, move_number, captures, return_captured=False):
        # It takes 40% less time than create_new_board_from_move (60% faster)
        enemy_position = None

        if move in self.get_possible_capture_moves(captures):
            if return_captured:
                enemy_position = self.perform_capture_move(move, move_number, captures, return_captured=return_captured)
            else:
                self.perform_capture_move(move, move_number, captures)
        else:
            self.perform_positional_move(move, move_number)

        if return_captured:
            return self, enemy_position
        else:
            return self

    def perform_capture_move(self, move, move_number, captures, return_captured=False):
        self.previous_move_was_capture = True
        piece = self.searcher.get_piece_by_position(move[0])
        originally_was_king = piece.king
        enemy_piece = piece.capture_move_enemies[move[1]]
        enemy_position = enemy_piece.position
        enemy_piece.capture()
        self.move_piece(move, move_number)
        if not originally_was_king and piece.king and (self.variant == 'english' or self.variant == 'italian'):
            further_capture_moves_for_piece = []
        elif not originally_was_king and self.variant != 'russian':
            was_king = piece.king
            piece.king = False
            further_capture_moves_for_piece = [capture_move for capture_move in self.get_possible_capture_moves(captures + [enemy_position]) if move[1] == capture_move[0]]
            if not further_capture_moves_for_piece and was_king:
                piece.king = True
        else:
            further_capture_moves_for_piece = [capture_move for capture_move in self.get_possible_capture_moves(captures + [enemy_position]) if move[1] == capture_move[0]]

        if further_capture_moves_for_piece:
            self.piece_requiring_further_capture_moves = self.searcher.get_piece_by_position(move[1])
        else:
            self.piece_requiring_further_capture_moves = None
            self.switch_turn()
        if return_captured:
            return enemy_position

    def perform_positional_move(self, move, move_number):
        self.previous_move_was_capture = False
        self.move_piece(move, move_number)
        self.switch_turn()

    def switch_turn(self):
        self.player_turn = BLACK if self.player_turn == WHITE else WHITE

    def move_piece(self, move, move_number):
        self.searcher.get_piece_by_position(move[0]).move(move[1], move_number)
        self.pieces = sorted(self.pieces, key=lambda piece: piece.position if piece.position else 0)

    def is_valid_row_and_column(self, row, column):
        if row < 0 or row >= self.height:
            return False

        if column < 0 or column >= self.width:
            return False

        return True

    def __setattr__(self, name, value):
        super(Board, self).__setattr__(name, value)

        if name == 'pieces':
            [piece.reset_for_new_board() for piece in self.pieces]

            self.searcher.build(self)



from draughts.engines.checkerboard_extra.engine_64 import Engine64
from draughts.engines.checkerboard_extra.engine_client import Engine32
import os
import draughts
import draughts.engine


class CheckerBoardEngine:
    def __init__(self, command, divide_time_by=40, checkerboard_timing=False, ENGINE=5):
        self.cwd = os.path.realpath(os.path.expanduser("."))
        if type(command) == str:
            command = [command]
        command = list(filter(bool, command))
        command = " ".join(command)
        self.command = os.path.join(self.cwd, command)
        self.ENGINE = ENGINE
        self.engine = None
        self.info = None
        self.id = {}
        self.result = None
        self.divide_time_by = divide_time_by
        self.checkerboard_timing = checkerboard_timing
        self._sent_variant = False
        self.bits = self._open_engine()
        self.id["name"] = self.engine.enginecommand('name')[0].decode()

    def _open_engine(self):
        try:
            self.engine = Engine64(self.command, self.cwd)
            return 64
        except Exception:
            self.engine = Engine32(self.command)  # I will see what I will do with it
            return 32

    def setoption(self, name, value):
        if name == 'divide-time-by':
            self.divide_time_by = value
        else:
            self.engine.enginecommand(f"set {name} {value}")

    def kill_process(self):
        if self.bits == 32:
            self.engine.shutdown_server32()
        else:
            self.engine.kill_process()

    def play(self, board, time_limit):
        time = time_limit.time
        inc = time_limit.inc
        depth = time_limit.depth
        nodes = time_limit.nodes
        movetime = time_limit.movetime

        if not inc:
            inc = 0

        if not self._sent_variant:
            if board.variant == 'russian':
                self.engine.enginecommand('set gametype 25')
            elif board.variant == 'brazilian':
                self.engine.enginecommand('set gametype 26')
            elif board.variant == 'italian':
                self.engine.enginecommand('set gametype 22')
            elif board.variant == 'english':
                self.engine.enginecommand('set gametype 21')
            self._sent_variant = True

        if board.move_stack:
            gamehist = f'set gamehist {board.last_non_reversible_fen} {" ".join(list(map(lambda move: move.pdn_move, board.non_reversible_moves)))}'
            if len(gamehist) > 256:
                gamehist = " ".join(gamehist[:256].split()[:-1])
            self.engine.enginecommand(gamehist)

        time_to_use = None
        if time:

            if time < 0 and inc < 0:
                time = 0
                inc = 0
            elif time < 0:
                inc = max(inc + time, 0)
                time = 0
            elif inc < 0:
                time = max(time + inc, 0)
                inc = 0

            if self.checkerboard_timing:
                if time < inc * .4:
                    time_to_use = inc * .4
                elif time < inc:
                    time_to_use = time
                else:
                    time_to_use = inc + (time - inc) / 2.5
            else:
                time_to_use = time / self.divide_time_by

        hub_pos_move, info, cbmove, result = self.engine.getmove(board, time_to_use, time, inc, movetime)

        if hub_pos_move:
            bestmove = draughts.Move(board, hub_position_move=hub_pos_move)
        else:
            steps = []
            positions = [cbmove['from']]
            jumps = max(cbmove['jumps'], 1)
            for pos in cbmove['path'][1:jumps]:
                positions.append(pos)
            positions.append(cbmove['to'])
            for pos in positions:
                steps.append(
                    self._row_col_to_num(board, pos[1], pos[0]))  # Checkerboard returns first the column, then the row

            bestmove = draughts.Move(board, steps_move=steps)

        self.info = info.decode()
        self.result = result
        return draughts.engine.PlayResult(bestmove, None, {'info': self.info, 'result': self.result})

    def _row_col_to_num(self, board, row, col):
        if row % 2 == 0:
            col = ((col + 2) / 2) - 1
        else:
            col = ((col + 1) / 2) - 1
        row = (board.board.height - 1) - row
        loc = board.board.position_layout.get(row, {}).get(col)
        return loc



# This file is an adaptation of draughtsnet (https://github.com/RoepStoep/draughtsnet) by RoepStoep.
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import os
import signal
import logging
import threading
import time
import draughts
import draughts.engine

logger = logging.getLogger(__name__)


class HubEngine:
    def __init__(self, command, ENGINE=5):
        if type(command) == str:
            command = [command]
        self.ENGINE = ENGINE
        self.info = {}
        self.id = {}
        cwd = os.path.realpath(os.path.expanduser("."))
        command = list(filter(bool, command))
        command = " ".join(command)
        self.p = self._open_process(command, cwd)
        self._last_sent = ""
        self.hub()

    def _open_process(self, command, cwd=None, shell=True, _popen_lock=threading.Lock()):
        kwargs = {
            "shell": shell,
            "stdout": subprocess.PIPE,
            "stderr": subprocess.STDOUT,
            "stdin": subprocess.PIPE,
            "bufsize": 1,  # Line buffered
            "universal_newlines": True,
        }

        if cwd is not None:
            kwargs["cwd"] = cwd

        # Prevent signal propagation from parent process
        try:
            # Windows
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        except AttributeError:
            # Unix
            kwargs["preexec_fn"] = os.setpgrp

        with _popen_lock:  # Work around Python 2 Popen race condition
            return subprocess.Popen(command, **kwargs)

    def kill_process(self):
        try:
            # Windows
            self.p.send_signal(signal.CTRL_BREAK_EVENT)
        except AttributeError:
            # Unix
            os.killpg(self.p.pid, signal.SIGKILL)

        self.p.communicate()

    def send(self, line):
        if line == "ponder-hit":
            while self._last_sent != "go ponder":
                pass
        logging.debug(f"{self.ENGINE} %s << %s {self.p.pid} {line}")

        self.p.stdin.write(line + "\n")
        self.p.stdin.flush()
        self._last_sent = line

    def recv(self):
        while True:
            line = self.p.stdout.readline()
            if line == "":
                raise EOFError()

            line = line.rstrip()

            logging.debug(f"{self.ENGINE} %s >> %s {self.p.pid} {line}")

            if line:
                return line

    def recv_hub(self):
        command_and_args = self.recv().split(None, 1)
        if len(command_and_args) == 1:
            return command_and_args[0], ""
        elif len(command_and_args) == 2:
            return command_and_args

    def hub(self):
        self.send("hub")

        engine_info = {}
        variants = set()

        while True:
            command, arg = self.recv_hub()

            if command == "wait":
                return engine_info, variants
            elif command == "id":
                parts = arg.split()
                i = 0
                while i < len(parts):
                    name_and_value = parts[i].split("=", 1)
                    if len(name_and_value) == 2:
                        value = name_and_value[1]
                        if value.startswith("\""):
                            value = value[1:]
                            i += 1
                            while i < len(parts) and not parts[i].endswith("\""):
                                value += " " + parts[i]
                                i += 1
                            value += " " + parts[i][:-1]
                        engine_info[name_and_value[0]] = value
                    i += 1
            elif command == "param":
                if arg.startswith("name=variant") and arg.find("type=enum") != -1:
                    argvalues = arg.split("=")[-1].replace("\"", "")
                    for variant in argvalues.split():
                        variants.add(variant)
            else:
                logging.warning("Unexpected engine response to hub: %s %s", command, arg)
            self.variants = variants
            self.id = engine_info

    def init(self):
        self.send("init")
        while True:
            command, arg = self.recv_hub()
            if command == "ready":
                break
            elif command == "init":
                pass
            else:
                logging.warning("Unexpected engine response to init: %s %s", command, arg)

    def ping(self):
        self.send("ping")
        while True:
            command, arg = self.recv_hub()
            if command == "pong":
                break
            else:
                logging.warning("Unexpected engine response to ping: %s %s", command, arg)

    def setoption(self, name, value):
        if value is True:
            value = "true"
        elif value is False:
            value = "false"
        elif value is None:
            value = "none"

        if name == 'variant' and self.variants or name != 'variant':
            self.send("set-param name=%s value=%s" % (name, value))

    def go(self, position, my_time=None, inc=None, movetime=None, clock=None, depth=None, analysisnodes=None, 
           handicap=None, searchnodes=None, ply=None, moves=None, ponder=False):
        if moves is None:
            moves = ''
        if moves and len(moves) != 0:
            self.send("pos pos=%s moves=\"%s\"" % (position, moves))
        else:
            self.send("pos pos=%s" % position)

        if handicap is not None and handicap > 0:
            self.send("level handicap=%s" % str(handicap))
        if ply is not None and ply > 0:
            self.send("level ply=%s" % str(ply))
        if searchnodes is not None and searchnodes > 0:
            self.send("level nodes=%s" % str(searchnodes))

        if movetime is not None and clock is not None:
            if position[0] == 'B':
                timeleft = clock["btime"] / 100.0
            else:
                timeleft = clock["wtime"] / 100.0
            increment = clock["inc"]
            if increment == 0 and timeleft < 50.0:
                movetime *= (timeleft / 50.0)
            elif increment == 1 and timeleft < 6.0:
                movetime *= (timeleft / 6.0)

        # if analysisnodes is not None and movetime is None:
        #     movetime = 4  # If no movetime is specified, allow 4 seconds to reach the required amount of nodes

        if depth is not None and depth > 0:
            self.send("level depth=%s" % str(depth))
        if movetime is not None:
            self.send("level move-time=%s" % str(movetime))

        if depth is None and movetime is None and clock is None and searchnodes is None:
            if inc:
                my_time -= inc  # Hub engines first add the increment
                self.send("level time={} inc={}".format(my_time, inc))
            else:
                self.send("level time={}".format(my_time))

        if analysisnodes is not None:
            self.send("level nodes=%s" % str(analysisnodes))
        if ponder:
            self.send("go ponder")
        else:
            self.send("go think")

        info = {}
        info["bestmove"] = None
        info["pondermove"] = None
        info["taken"] = ""
        info["pondertaken"] = ""

        start = time.time()
        while True:
            command, arg = self.recv_hub()
            arg_split = arg.split()

            forcestop = False
            if command == "done":
                if len(arg) == 0:
                    # info["bestmove"] = ""
                    # info["taken"] = ""
                    # info["score"] = {"win": 0}
                    # return info
                    return None, None, None, None
                bestmoveval = arg_split[0]
                if bestmoveval and bestmoveval.find("=") != -1:
                    bestmove = bestmoveval.split("=")[1]
                    if bestmove.find("x") != -1:
                        fields = bestmove.split("x")
                        origdest = fields[:2]
                        taken = fields[2:]
                    else:
                        origdest = bestmove.split("-")[:2]
                        taken = []
                    if len(origdest) == 2:
                        info["bestmove"] = "%02d%02d" % (int(origdest[0]), int(origdest[1]))
                    alltaken = ""
                    for t in taken:
                        alltaken += "%02d" % int(t)
                    info["taken"] = alltaken
                if len(arg_split) >= 2:
                    pondermoveval = arg_split[1]
                    if pondermoveval and pondermoveval.find("=") != -1:
                        pondermove = pondermoveval.split("=")[1]
                        if pondermove.find("x") != -1:
                            fields = pondermove.split("x")
                            origdest = fields[:2]
                            taken = fields[2:]
                        else:
                            origdest = pondermove.split("-")[:2]
                            taken = []
                        if len(origdest) == 2:
                            info["pondermove"] = "%02d%02d" % (int(origdest[0]), int(origdest[1]))
                        alltaken = ""
                        for t in taken:
                            alltaken += "%02d" % int(t)
                        info["pondertaken"] = alltaken
                return info["bestmove"], info["pondermove"], info["taken"], info["pondertaken"]
            elif command == "info":
                arg = arg or ""

                # Parse all other parameters
                cur_nodes = 0
                cur_depth = 0
                parts = arg.split()
                i = 0
                while i < len(parts):
                    name_and_value = parts[i].split("=", 1)
                    if len(name_and_value) == 2:
                        value = name_and_value[1]
                        if value.startswith("\""):
                            value = value[1:]
                            i += 1
                            while i < len(parts) and not parts[i].endswith("\""):
                                value += " " + parts[i]
                                i += 1
                            value += " " + parts[i][:-1]
                        if name_and_value[0] in ["nodes", "depth"]:
                            info[name_and_value[0]] = int(value)
                        elif name_and_value[0] == "score":
                            score = int(float(value) * 100)
                            if abs(score) > 9000:
                                if score > 0:
                                    plies = 10000 - score
                                    moves = int((plies + plies % 2) / 2)
                                else:
                                    plies = -(10000 + score)
                                    moves = int((plies - plies % 2) / 2)
                                info["score"] = {"win": moves}
                            elif abs(score) > 8000:
                                if score > 0:
                                    plies = 9000 - score
                                    moves = int((plies + plies % 2) / 2)
                                else:
                                    plies = -(9000 + score)
                                    moves = int((plies - plies % 2) / 2)
                                info["score"] = {"win": moves}
                            else:
                                info["score"] = {"cp": score}
                        elif name_and_value[0] == "time":
                            info["time"] = int(float(value) * 1000)
                        elif name_and_value[0] == "nps":
                            info["nps"] = round(float(value))
                        elif name_and_value[0] != "mean-depth":
                            info[name_and_value[0]] = value
                        if name_and_value[0] == "nodes":
                            cur_nodes = int(value)
                        if name_and_value[0] == "depth":
                            cur_depth = int(value)
                    i += 1
                    self.info = info

                # Check if we have reached the node count for analysis
                if analysisnodes is not None and (cur_nodes > analysisnodes or cur_depth > 50):
                    forcestop = True
            else:
                logging.warning("Unexpected engine response to go: %s %s", command, arg)

            # Force play if movetime is exceeded (not exact as we only reach this point after the engine produces output)
            if forcestop or (movetime is not None and time.time() - start > movetime):
                self.send("stop")

    def known_variant(self, variant):
        return self._parse_variant(variant) in ["normal", "bt", "frisian", "losing"]

    def _parse_variant(self, variant):
        variant = variant.lower()

        if variant in ["standard", "fromposition"]:
            return "normal"
        elif variant == "breakthrough":
            return "bt"
        elif variant == "antidraughts":
            return "losing"
        elif variant == "frysk":
            return "frisian"
        else:
            return variant

    def stop(self):
        self.send("stop")

    def ponderhit(self):
        self.send("ponder-hit")

    def quit(self):
        self.send("quit")

    def play(self, board, time_limit, ponder):
        time = time_limit.time
        inc = time_limit.inc
        depth = time_limit.depth
        nodes = time_limit.nodes
        movetime = time_limit.movetime
        hub_moves = board.move_stack
        hub_moves = list(map(lambda move: move.hub_move, hub_moves))
        bestmove, pondermove, taken, pondertaken = self.go(board.initial_hub_fen, moves=' '.join(hub_moves), my_time=time, inc=inc, depth=depth, analysisnodes=nodes, movetime=movetime, ponder=ponder)

        ponder_move = None
        if bestmove is None:
            return None, None
        ponder_board = board.copy()
        best_move = draughts.Move(ponder_board, hub_position_move=bestmove + taken)
        if pondermove:
            for move in best_move.board_move:
                ponder_board.move(move)
            ponder_move = draughts.Move(ponder_board, hub_position_move=pondermove + pondertaken)

        return draughts.engine.PlayResult(best_move, ponder_move, self.info)



import os

from msl.loadlib import Client64


class Engine32(Client64):
    def __init__(self, command):

        command = command.replace('\\', '\\\\')
        with open(os.path.join(os.path.dirname(__file__), 'engine_name.txt'), 'w') as file:
            file.write(command)

        super(Engine32, self).__init__(module32='engine_server', append_sys_path=os.path.dirname(__file__))

    def enginecommand(self, command):
        return self.request32('enginecommand', command)

    def getmove(self, game, maxtime=None, time=None, increment=None, movetime=None):
        assert maxtime is not None or time is not None or movetime is not None
        return self.request32('getmove', game, maxtime, time, increment, movetime)



from math import ceil

WHITE = 2
BLACK = 1


class Piece:

    def __init__(self, variant='standard'):
        self.player = None
        self.other_player = None
        self.king = False
        self.captured = False
        self.position = None
        self.board = None
        self.became_king = -100
        self.capture_move_enemies = {}
        self.variant = variant
        self.reset_for_new_board()

    def reset_for_new_board(self):
        self.possible_capture_moves = None
        self.possible_positional_moves = None

    def is_movable(self, captures):
        return (self.get_possible_capture_moves(captures) or self.get_possible_positional_moves()) and not self.captured

    def capture(self):
        self.captured = True
        self.position = None

    def move(self, new_position, move_number):
        self.position = new_position
        was_king = self.king
        self.king = self.king or self.is_on_enemy_home_row()
        if self.king != was_king:
            self.became_king = move_number

    def get_possible_capture_moves(self, captures):
        if self.possible_capture_moves is None:
            self.possible_capture_moves = self.build_possible_capture_moves(captures)

        return self.possible_capture_moves

    def build_possible_capture_moves(self, captures):
        adjacent_enemy_positions = list(filter((lambda position: position in self.board.searcher.get_positions_by_player(self.other_player)), self.get_adjacent_positions(capture=True)))
        capture_move_positions = []
        for enemy_position in adjacent_enemy_positions:
            enemy_piece = self.board.searcher.get_piece_by_position(enemy_position)
            if self.variant == 'italian' and not self.king and enemy_piece.king:
                # Men can't capture kings in italian draughts
                continue
            positions_behind_enemy = self.get_position_behind_enemy(enemy_piece, captures)
            for position_behind_enemy in positions_behind_enemy:

                if (position_behind_enemy is not None) and self.board.position_is_open(position_behind_enemy):
                    capture_move_positions.append(position_behind_enemy)
                    self.capture_move_enemies[position_behind_enemy] = enemy_piece

        return self.create_moves_from_new_positions(capture_move_positions)

    def get_position_behind_enemy(self, enemy_piece, captures):
        """
        Gets the positions the piece can land after capturing the enemy piece.
        :param captures: it contains all the positions of the pieces that this piece has captured during this multi-capture,
        because kings in multi-captures can't go over a piece they have captured in that move sequence in most variants
        """
        if not self.king:
            current_row = self.get_row()
            current_column = self.get_column()
            enemy_column = enemy_piece.get_column()
            enemy_row = enemy_piece.get_row()
            if self.variant in ['frisian', 'frysk!', 'turkish']:  # Orthogonal captures
                row_difference = current_row - enemy_row
                column_difference = current_column - enemy_column
                if (column_difference == 0 or row_difference == 0) and (row_difference % 2 == 0 if self.variant in ['frisian', 'frysk!'] else True):
                    next_row = enemy_row - row_difference
                    next_column = enemy_column - column_difference
                    return [self.board.position_layout.get(next_row, {}).get(next_column)]

            column_adjustment = -1 if current_row % 2 == 0 else 1
            column_behind_enemy = current_column + column_adjustment if current_column == enemy_column else enemy_column
            row_behind_enemy = enemy_row + (enemy_row - current_row)

            return [self.board.position_layout.get(row_behind_enemy, {}).get(column_behind_enemy)]
        else:
            positions = []
            current_row = self.get_row()
            current_column = self.get_column()
            enemy_column = enemy_piece.get_column()
            enemy_row = enemy_piece.get_row()
            if self.variant in ['frisian', 'frysk!', 'turkish']:  # Orthogonal captures
                positions_to_check = []
                row_difference = current_row - enemy_row
                column_difference = current_column - enemy_column
                add_row = 0
                if row_difference < 0:
                    add_row = -1
                elif row_difference > 0:
                    add_row = 1
                add_column = 0
                if column_difference < 0:
                    add_column = -1
                elif column_difference > 0:
                    add_column = 1
                if (column_difference == 0 or row_difference == 0) and (row_difference % 2 == 0 if self.variant in ['frisian', 'frysk!'] else True):
                    for multiplier in range(1, max(self.board.width, self.board.height)):
                        if multiplier % 2 == 1 and self.variant in ['frisian', 'frysk!']:
                            continue
                        next_row = current_row - multiplier * add_row
                        next_column = current_column - multiplier * add_column
                        if next_row in self.board.position_layout and next_column in self.board.position_layout[next_row]:
                            positions_to_check.append(self.board.position_layout.get(next_row, {}).get(next_column))
                        next_row = enemy_row - multiplier * add_row
                        next_column = enemy_column - multiplier * add_column
                        if next_row in self.board.position_layout and next_column in self.board.position_layout[next_row]:
                            positions.append(self.board.position_layout.get(next_row, {}).get(next_column))

                    new_positions = []
                    for index, position in enumerate(positions_to_check):
                        enemy_piece_found = False
                        for semi_position in positions_to_check[:index + 1]:
                            piece = self.board.searcher.get_piece_by_position(semi_position)
                            if piece is not None:
                                if piece.player == self.player or enemy_piece_found:  # It stops if it meets a piece of the same color or another opponent piece
                                    break
                                else:
                                    enemy_piece_found = True
                            elif semi_position in ([captures[-1]] if self.variant == 'turkish' and captures else captures):
                                # Kings in multi-captures can't go over a piece they have captured in that move sequence in frisian and frysk!
                                # In turkish they can but we use the last capture, so we prevent the piece from turning 180 degrees, which is not allowed
                                break
                        else:
                            if position in positions:
                                new_positions.append(position)
                            continue
                        break
                    positions = new_positions

                    return positions
            elif self.variant == 'english' or self.variant == 'italian':
                # Same as if it is not a king
                column_adjustment = -1 if current_row % 2 == 0 else 1
                column_behind_enemy = current_column + column_adjustment if current_column == enemy_column else enemy_column
                row_behind_enemy = enemy_row + (enemy_row - current_row)

                return [self.board.position_layout.get(row_behind_enemy, {}).get(column_behind_enemy)]
            adjacent_positions = self.get_adjacent_positions(capture=True)
            down_direction = enemy_row > current_row
            left_direction = enemy_column < current_column if current_row % 2 == 1 else enemy_column <= current_column

            legal_adjacent_positions = []
            was_king = enemy_piece.king
            enemy_piece.king = True
            for position in enemy_piece.get_adjacent_positions(capture=True):
                column = (position - 1) % self.board.width
                row = self.get_row_from_position(position)
                down_direction_possible = row > enemy_row
                left_direction_possible = column < enemy_column if enemy_row % 2 == 1 else column <= enemy_column
                # Checks if the captured piece square is between the starting square and the landing square
                # For example, if the captured piece square is to the left and down of the starting square, we check if the landing square is to the left and down of the captured piece square
                # It also checks if the position is in a pseudolegal list of legal moves (because we can have landing squares that aren't in the same diagonal as the staring square (e.g. if the piece is in 28, 31 isn't a possible landing square))
                if down_direction_possible == down_direction and left_direction_possible == left_direction and position in adjacent_positions:
                    legal_adjacent_positions.append(self.board.position_layout.get(row, {}).get(column))
            enemy_piece.king = was_king
            adjacent_positions = legal_adjacent_positions

            positions_to_check = []
            row_to_check = current_row
            position_to_check = self.position
            subtract_for_variant = 1 if self.variant == 'brazilian' or self.variant == 'russian' else 0
            for add in range(1, self.board.height):
                if down_direction and not left_direction and row_to_check % 2 == 1:
                    position_to_check += 5 - subtract_for_variant
                    row_to_check += 1
                elif down_direction and not left_direction and row_to_check % 2 == 0:
                    position_to_check += 6 - subtract_for_variant
                    row_to_check += 1
                elif down_direction and left_direction and row_to_check % 2 == 1:
                    position_to_check += 4 - subtract_for_variant
                    row_to_check += 1
                elif down_direction and left_direction and row_to_check % 2 == 0:
                    position_to_check += 5 - subtract_for_variant
                    row_to_check += 1
                elif not down_direction and not left_direction and row_to_check % 2 == 1:
                    position_to_check -= 5 - subtract_for_variant
                    row_to_check -= 1
                elif not down_direction and not left_direction and row_to_check % 2 == 0:
                    position_to_check -= 4 - subtract_for_variant
                    row_to_check -= 1
                elif not down_direction and left_direction and row_to_check % 2 == 1:
                    position_to_check -= 6 - subtract_for_variant
                    row_to_check -= 1
                elif not down_direction and left_direction and row_to_check % 2 == 0:
                    position_to_check -= 5 - subtract_for_variant
                    row_to_check -= 1
                if self.get_row_from_position(position_to_check) == row_to_check:
                    positions_to_check.append(position_to_check)
                else:
                    break

            for index, position in enumerate(positions_to_check):  # Checks if the piece will encounter any other piece in its path
                enemy_piece_found = False
                for semi_position in positions_to_check[:index + 1]:
                    piece = self.board.searcher.get_piece_by_position(semi_position)
                    if piece is not None:
                        if piece.player == self.player or enemy_piece_found:  # It stops if it meets a piece of the same color or another opponent piece
                            break
                        else:
                            enemy_piece_found = True
                    elif semi_position in captures:  # Kings in multi-captures can't go over a piece they have captured in that move sequence in most variants
                        break
                else:
                    if position in adjacent_positions:
                        positions.append(position)
                    continue
                break
            return positions

    def get_possible_positional_moves(self):
        if self.possible_positional_moves is None:
            self.possible_positional_moves = self.build_possible_positional_moves()

        return self.possible_positional_moves

    def build_possible_positional_moves(self):
        new_positions = list(filter((lambda position: self.board.position_is_open(position)), self.get_adjacent_positions()))

        return self.create_moves_from_new_positions(new_positions)

    def create_moves_from_new_positions(self, new_positions):
        return [[self.position, new_position] for new_position in new_positions]

    def get_adjacent_positions(self, capture=False):
        criteria = bool(self.king) if self.variant in ['english', 'italian', 'turkish'] else bool(capture or self.king)  # In english, italian and turkish men can't capture backwards
        return self.get_directional_adjacent_positions(forward=True, capture=capture) + (self.get_directional_adjacent_positions(forward=False, capture=capture) if criteria else [])

    def get_column(self):
        return (self.position - 1) % self.board.width

    def get_row(self):
        return self.get_row_from_position(self.position)

    def is_on_enemy_home_row(self):
        return self.get_row() == self.get_row_from_position(1 if self.other_player == BLACK else self.board.position_count)

    def get_row_from_position(self, position):
        return ceil(position / self.board.width) - 1

    def get_directional_adjacent_positions(self, forward, capture=False):
        # In frisian, forward=True includes left and up and forward=False includes right and down
        # e.g. If the piece is at 33, square 32 will be considered in forward=True and square 34 will be considered in forward=False
        # In turkish, if the piece is a king, forward=True includes left and up and forward=False includes right and down
        # if it isn't a king forward=True, calculates up, left and right
        if not self.king:
            if self.variant == 'turkish':
                positions = []
                current_row = self.get_row()
                current_column = self.get_column()
                next_row = current_row + (1 if self.player == BLACK else -1)
                if next_row in self.board.position_layout:
                    positions.append(self.board.position_layout[next_row][current_column])
                next_column = current_column - 1
                if next_column in self.board.position_layout[current_row]:
                    positions.append(self.board.position_layout[current_row][next_column])
                next_column = current_column + 1
                if next_column in self.board.position_layout[current_row]:
                    positions.append(self.board.position_layout[current_row][next_column])
                return positions
            if (self.variant == 'frisian' or self.variant == 'frysk!') and capture:
                positions = []
                current_row = self.get_row()
                current_column = self.get_column()
                next_row = current_row + ((2 if self.player == BLACK else -2) * (1 if forward else -1))
                next_column = current_column + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

                # Orthogonal squares (only in frisian and frysk!)
                if next_row not in self.board.position_layout:
                    pass
                else:
                    positions.append(self.board.position_layout[next_row][current_column])
                if next_column not in self.board.position_layout[current_row]:
                    pass
                else:
                    positions.append(self.board.position_layout[current_row][next_column])

                # Diagonal squares (as in standard draughts)
                next_row = current_row + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

                if next_row not in self.board.position_layout:
                    pass
                else:

                    next_column_indexes = self.get_next_column_indexes(current_row, self.get_column())

                    positions += [self.board.position_layout[next_row][column_index] for column_index in next_column_indexes]
                return positions
            else:
                current_row = self.get_row()
                next_row = current_row + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

                if next_row not in self.board.position_layout:
                    return []

                next_column_indexes = self.get_next_column_indexes(current_row, self.get_column())

                return [self.board.position_layout[next_row][column_index] for column_index in next_column_indexes]
        else:
            positions = []
            current_row = self.get_row()
            current_column = self.get_column()
            if self.variant == 'turkish':
                for multiplier in range(1, self.board.height):
                    next_row = current_row + multiplier * (1 if self.player == BLACK else -1) * (1 if forward else -1)
                    if next_row in self.board.position_layout:
                        positions.append(self.board.position_layout[next_row][current_column])
                for multiplier in range(1, self.board.width):
                    next_column = current_column + multiplier * (1 if self.player == BLACK else -1) * (1 if forward else -1)
                    if next_column in self.board.position_layout[current_row]:
                        positions.append(self.board.position_layout[current_row][next_column])
                return positions
            if (self.variant == 'frisian' or self.variant == 'frysk!') and capture:
                # Orthogonal squares
                for add_column in range(1, self.board.width):
                    next_column = current_column + ((add_column if self.player == BLACK else -add_column) * (1 if forward else -1))
                    if next_column not in self.board.position_layout[current_row]:
                        positions += []
                        continue
                    positions += [self.board.position_layout[current_row][next_column]]
                for add_row in range(2, self.board.height, 2):
                    next_row = current_row + ((add_row if self.player == BLACK else -add_row) * (1 if forward else -1))
                    if next_row not in self.board.position_layout:
                        positions += []
                        continue
                    positions += [self.board.position_layout[next_row][current_column]]
            elif self.variant == 'english' or self.variant == 'italian':
                # Same as if it is not a king
                next_row = current_row + ((1 if self.player == BLACK else -1) * (1 if forward else -1))

                if next_row not in self.board.position_layout:
                    return []

                next_column_indexes = self.get_next_column_indexes(current_row, self.get_column())

                return [self.board.position_layout[next_row][column_index] for column_index in next_column_indexes]

            # Diagonal squares
            positions_diagonal_1 = []
            positions_diagonal_2 = []
            for i in range(1, self.board.height):
                next_row = current_row + ((i if self.player == BLACK else -i) * (1 if forward else -1))

                if next_row not in self.board.position_layout:
                    positions += []
                    continue

                next_column_indexes = self.get_next_column_indexes(current_row, current_column, i, True)
                if 0 <= next_column_indexes[0] < self.board.width:
                    positions_diagonal_1.append(self.board.position_layout[next_row][next_column_indexes[0]])
                if 0 <= next_column_indexes[1] < self.board.width:
                    positions_diagonal_2.append(self.board.position_layout[next_row][next_column_indexes[1]])

            for index, position in enumerate(positions_diagonal_1):
                for semi_position in positions_diagonal_1[:index + 1]:
                    piece = self.board.searcher.get_piece_by_position(semi_position)
                    if piece is not None and not capture:
                        # If we encounter a piece, and we are searching for a positional move not a capture move, we break the loop
                        break
                else:
                    positions.append(position)
                    continue
                break

            for index, position in enumerate(positions_diagonal_2):
                for semi_position in positions_diagonal_2[:index + 1]:
                    piece = self.board.searcher.get_piece_by_position(semi_position)
                    if piece is not None and not capture:
                        # If we encounter a piece, and we are searching for a positional move not a capture move, we break the loop
                        break
                else:
                    positions.append(position)
                    continue
                break

            return positions

    def get_next_column_indexes(self, current_row, current_column, i=1, unfiltered=False):
        """
        Get the index of the next column.
        It isn't as simple as finding the next row (where we only add or subtract 1) but the column index only changes
        once every 2 rows.
        e.g. Square 32 and 27 are both in the second row, but 27 and 31 are not.
        :param i: current_row ± i is the row we should search (the columns for current_row + i are the same as current_row - i). e.g if current row=3 and i=2, we will return the 2 possible columns for row 5 (which will be the same for row 1)
        """
        column_indexes = [0, 0]
        start_right = True if current_row % 2 == 0 else False
        for semi_i in range(1, i + 1):
            if start_right and semi_i % 2 == 1 or not start_right and semi_i % 2 == 0:
                column_indexes[1] += 1
            else:
                column_indexes[0] -= 1

        column_indexes = list(map(lambda value: value + current_column, column_indexes))

        if unfiltered:
            return column_indexes
        else:
            return filter((lambda column_index: 0 <= column_index < self.board.width), column_indexes)

    def __setattr__(self, name, value):
        super(Piece, self).__setattr__(name, value)

        if name == 'player':
            self.other_player = BLACK if value == WHITE else WHITE
